@@ -148,6 +148,8 @@ static void xfrm_policy_timer(unsigned long data)
 	int warn = 0;
 	int dir;
 
+	read_lock(&xp->lock);
+
 	if (xp->dead)
 		goto out;
 
@@ -197,6 +199,7 @@ static void xfrm_policy_timer(unsigned long data)
 		xfrm_pol_hold(xp);
 
 out:
+	read_unlock(&xp->lock);
 	xfrm_pol_put(xp);
 	return;
 
@@ -361,7 +364,6 @@ int xfrm_policy_insert(int dir, struct xfrm_policy *policy, int excl)
 	if (delpol) {
 		xfrm_policy_kill(delpol);
 	}
-	wake_up(&km_waitq);
 	return 0;
 }
 
@@ -384,7 +386,6 @@ struct xfrm_policy *xfrm_policy_bysel(int dir, struct xfrm_selector *sel,
 	if (pol && delete) {
 		atomic_inc(&flow_cache_genid);
 		xfrm_policy_kill(pol);
-		wake_up(&km_waitq);
 	}
 	return pol;
 }
@@ -407,7 +408,6 @@ struct xfrm_policy *xfrm_policy_byid(int dir, u32 id, int delete)
 	if (pol && delete) {
 		atomic_inc(&flow_cache_genid);
 		xfrm_policy_kill(pol);
-		wake_up(&km_waitq);
 	}
 	return pol;
 }
@@ -430,7 +430,6 @@ void xfrm_policy_flush(void)
 	}
 	atomic_inc(&flow_cache_genid);
 	write_unlock_bh(&xfrm_policy_lock);
-	wake_up(&km_waitq);
 }
 
 int xfrm_policy_walk(int (*func)(struct xfrm_policy *, int, int, void*),
@@ -559,7 +558,6 @@ int xfrm_sk_policy_insert(struct sock *sk, int dir, struct xfrm_policy *pol)
 	if (old_pol) {
 		xfrm_policy_kill(old_pol);
 	}
-	wake_up(&km_waitq);
 	return 0;
 }
 
@@ -1019,6 +1017,8 @@ static int stale_bundle(struct dst_entry *dst)
 
 static void xfrm_dst_destroy(struct dst_entry *dst)
 {
+	if (!dst->xfrm)
+		return;
 	xfrm_state_put(dst->xfrm);
 	dst->xfrm = NULL;
 }
